;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; om-bits: Pieces of code taken from OpenMusic that 
;;; are required for simpleTree.lisp to work in CM-ETF
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:28>

;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


(in-package :om)

;;;;==============================================================
;;;; from kernel.lisp:

(defmethod om* ((self number) (num number))
  :initvals '(0 0) :indoc '("first input" "second input") :icon 195
  :doc "product of two numbers or trees."
  (* self num))

(defmethod om* ((self number) (num list))
  (mapcar #'(lambda (input)
              (om* self input)) num))

(defmethod om* ((self list) (num number))
  (mapcar #'(lambda (input)
              (om*  input num)) self))

(defmethod om* ((self list) (num list))
  (mapcar #'(lambda (input1 input2)
              (om* input1 input2)) self num))

;;;;==============================================================
;;;; from MusicUser-CspTechn.lisp:

(defvar *distance-function*)
(setf *distance-function* #'-)
(defvar *integrate-distance-fun*)
(setf *integrate-distance-fun* #'+)
(defvar *distance-neutral-element*)
(setf *distance-neutral-element* 0)

(defun x->dx (xs)
  "Returns the list of the intervals between the contiguous values of a list <xs>. 
For example
(x->dx '(0 4 5 9 6 2 3 3))
 will return  (4 1 4 -3 -4 1 0)  "
  (mapcar *distance-function* (cdr xs) xs))

(defun rest-abs (x y dist-fun) (abs (funcall dist-fun x y)))

(defun x->dx-abs (xs)
  "Returns the list of the intervals (absolutes) between the contiguous values of a list <xs>. 
For example
(x->dx '(0 4 5 9 6 2 3 3))
 will return  (4 1 4 -3 -4 1 0)  "
  (mapcar #'(lambda (x y) (rest-abs x y *distance-function*)) (cdr xs) xs))

(defun dx->x (start dxs)
 (let ((x start))
    (cons x (mapcar #'(lambda (dx) (setf x (funcall *integrate-distance-fun*  x dx)) x) dxs))))

(defun all-lists (i j  &optional (step 1) (level 0) (alt 0)  &key (test-end #'>=))
  (let (res)
    (do ((x 0 (+ step x))) ((funcall test-end x (- j i)) (nreverse res))
      (push (list (+ x i) (if (zerop (rem x 2)) level (+ level alt))) res ))))

(defun all-lists-2 (i j  &optional (step 1) (level 0) (alt 0)  )
  (let (res)
    (do ((x 0 (+ step x))) ((> x (- j i)) (nreverse res))
      (push (list (+ x i) (if (zerop (rem x 2)) level (+ level alt))) res ))))
;;(all-lists-2   2 8)

;;;;==============================================================
;;;; from Mcl2Cmucl.lisp

(defmacro while (test &body body)
  (let ((testlab (gensym))
        (toplab (gensym)))
    `(tagbody
      (go ,testlab)
      ,toplab
      (progn ,@body)
      ,testlab
      (when ,test (go ,toplab)))))

;;;;==============================================================
;;;; from LispTools.lisp

(defun list! (thing) (if (listp thing) thing (list thing)))

(defmacro newl (lst elem) `(push ,elem ,lst))

(defmacro nextl (lst &optional symb)
  (if symb
      `(setq ,symb (pop ,lst))
    `(pop ,lst) ))

;;;;==============================================================
;;;; from lists.lisp:

(defun rev-flat (lst)  
  (let ((l ()))
    (while lst
      (if (not (consp (car lst)))
	  (newl l (nextl lst))
	(setq l (nconc (rev-flat (nextl lst)) l))))
    l ))

(defun lo-flat (list) 
  (cond ((atom list) list)
        ((atom (car list)) (cons (car list) (lo-flat (cdr list))))
        ((atom (caar list)) (apply 'append list))
        (t (cons (lo-flat (car list)) (lo-flat (cdr list))))))

(defmethod flat-low ((list list)) 
  (lo-flat list))

(defmethod flat-once ((list list))
  (if (consp (car list))
      (apply 'append list)  list))

(defmethod flat-one ((list list))
  (loop for item in list
        append (list! item)))

(defmethod n-flat-one ((list list) (level integer))
  (let ((rep list))
    (loop for i from 1 to level do
          (setf rep (flat-one rep)))
    rep))

(defmethod flat ((lst list) &optional (level nil)) 
  ;; changed from defmethod* - NK
  :initvals '(nil nil)
  :indoc '("one list" "level of parenthesis")
  :icon 235
  :doc "Transforms tree-lists into flat lists.
If <level> is 1 (resp n) remove 1 (rep. n) level(s) of list imbrication.
If <level> is NIl (default) remove all levels of imbrication, down to a purely flat list."
  (cond
   ((null level) (nreverse (rev-flat lst)))
   ((= level 0) lst)
   ((and (integerp level) (plusp level))
    (n-flat-one lst level))
   (t lst)))
