;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

;Modified for linux by: Gerardo M. Sarria M. and Jose Fernando Diago
;                       Pontificia Universidad Javeriana - Cali, Colombia

;DocFile
;
;Last Modifications:
;18/10/1997 First date.
;03/07/2006 Changes by Niklas Kambeitz for CM-ETF:
;  - fixed parentheses in BETTER-PREDEFINED-SUBDIV?
;  - changed om::defmethod* to defmethod (MKTREE)
;  - changed om::defmethod! to defmethod (TREE2RATIO)
;
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;*****************************************************************************
;; This code was written 0rjan Sandred, IRCAM 03/1999

;; Use the function simple->tree. All other functions are called from this one.
;; This function builds a hierarchical rhythmtree from a sequence of rhythm
;; values on the format (1/4 1/4 1/4 1/4).

;; In "better-predefined-subdiv?" you can define other way of notation
;; for special cases (choise of subtrees).
;;*****************************************************************************

(in-package :om) 

;;-----------------------------------------------------------------------------

(defun make-proportional-cell (dur-list)
  (mapcar #'(lambda (dur) (* dur (apply 'lcm (mapcar 'denominator dur-list))))
          dur-list))

;;offset from 1 always!!!
;;Because otherwise the first pause ( -0!) will dissapear.
(defun x-dx-pause-ok (x-list)
  (mapcar '*
          (mapcar #'(lambda (absdur) (if (> absdur 0) 1 -1)) x-list)
          (om::x->dx (mapcar 'abs x-list))))

(defun dx-x-pause-ok (starttime x-list)
  (mapcar '*
          (append (mapcar #'(lambda (absdur)
			      (if (> absdur 0) 1 -1)) x-list) '(1))
          (om::dx->x starttime (mapcar 'abs x-list))))

(defun simplify-proportions (proportion-list)
  (mapcar #'(lambda (value)
	      (/ value (apply 'gcd proportion-list)))
	  proportion-list))

;;all negative numbers following eachother should be fused to single negative numbers
(defun fuse-pauses (proportional-list)
  (let ((pointer-to-list 0))
    (simplify-proportions
     (loop until (>= pointer-to-list (length proportional-list))
           collect (if (< (nth pointer-to-list proportional-list) 0)
		       (let ((this-pause (nth pointer-to-list proportional-list)))
			 (incf pointer-to-list)
			 (loop until (or (>= pointer-to-list (length proportional-list))
					 (> (nth pointer-to-list proportional-list) 0))
			       do (progn
				    (incf this-pause (nth pointer-to-list proportional-list))
				    (incf pointer-to-list)))
			 this-pause)
                     (progn (incf pointer-to-list)
                            (nth (1- pointer-to-list) proportional-list)))))))

;;if a pause first, start list with -1
;;if a tied note first, start list on next event (i.e. /= 1)
(defun make-sub-tree (local-onset)
  (list 1
        ;;(fuse-pauses
	(cond ((and (or (= (car local-onset) 1) (= (car local-onset) -1)))
	       (make-proportional-cell (x-dx-pause-ok local-onset)))
	      (t
	       (let ((proportional-list
		      (make-proportional-cell (x-dx-pause-ok (cons 1 local-onset)))))
		 (cons (float (first proportional-list))
		       (cdr proportional-list)))))))

(defun filter-events-between (start stop onsettimes)
  (let ((no-low-values (member start onsettimes
			       :test #'(lambda (item value)
					 (<= item (abs value))))))
    (reverse (member stop (reverse no-low-values)
		     :test #'(lambda (item value) (>= item (abs value)))))))

(defun build-local-times (global-onsets global-start)
  (mapcar #'(lambda (onset) (if (> onset 0)
				(- onset (1- global-start))
			      (+ onset (1- global-start))))
	  global-onsets))

(defun create-beat (global-onset global-start beat-length)
  (let ((local-onset (build-local-times global-onset global-start))
        tree)
    (if (not local-onset) (setf local-onset (list (1+ beat-length))))
    (if (= (car (last local-onset)) (- -1 beat-length))
	(setf local-onset (append (butlast local-onset)
				  (list (1+ beat-length)))))
    (if (/= (car (last local-onset)) (1+ beat-length))
	(setf local-onset (append local-onset (list (1+ beat-length)))))
    (setf tree (better-predefined-subdiv? (make-sub-tree local-onset)))
    (if (= (length (cadr tree)) 1)
	(caadr tree)
      tree)))
    
(defun fuse-pauses-and-tied-notes-between-beats (measure-tree no-of-beats)
  (let ((beat-nr 0))
    (loop until (>=  beat-nr no-of-beats)
          collect (cond ((and (typep (nth beat-nr measure-tree) 'number)
                              (< (nth beat-nr measure-tree) 0))
                         (let ((value (nth beat-nr measure-tree)))
                           (incf beat-nr)
                           (loop until (or (typep (nth beat-nr measure-tree)
						  'list)
                                           (>=  beat-nr no-of-beats)
                                           (and (typep (nth beat-nr measure-tree) 'integer)
                                                (> (nth beat-nr measure-tree) 0)))
                                 do (progn
				      (decf value (truncate (abs (nth beat-nr measure-tree))))
				      (incf beat-nr)))
                           value))
                        ((and (typep (nth beat-nr measure-tree) 'number))
                         (let ((value (nth beat-nr measure-tree)))
                           (incf beat-nr)
                           (loop until (not (typep (nth beat-nr measure-tree) 'float))
                                 do (progn
				      (incf value (truncate (nth beat-nr measure-tree)))
				      (incf beat-nr)))
                           value))
                        (t (incf beat-nr)
                           (nth (1- beat-nr) measure-tree))))))

(defun build-one-measure (local-onset no-of-beats beat-length)
  (let ((beatlist (om::dx->x 1 (make-list no-of-beats
					  :initial-element beat-length)))
        tree)
    (setf tree
          (fuse-pauses-and-tied-notes-between-beats
           (loop for beat-nr from 0 to (- (length beatlist) 2)
                 collect (let ((these-events (filter-events-between (nth beat-nr beatlist) (nth (1+ beat-nr) beatlist) local-onset)))
                           ;;check if tied pause within subtree - if yes: give startpoint as pause
                           (if (and these-events
                                    (/= (abs (first these-events)) (nth beat-nr beatlist))
                                    (get-onsettime-before (nth beat-nr beatlist) local-onset)
                                    (> 0 (get-onsettime-before (nth beat-nr beatlist) local-onset)))
			       (setf these-events (append (list (- 0 (nth beat-nr beatlist)))
							  these-events))) 
                           ;;check if tied pause within subtree - if yes: give startpoint as pause
                           (if (and (not these-events)
                                    (get-onsettime-before (nth beat-nr beatlist) local-onset)
                                    (> 0 (get-onsettime-before (nth beat-nr beatlist) local-onset)))
			       (setf these-events (list (- 0 (nth beat-nr beatlist)))))
                           
                           (create-beat these-events (nth beat-nr beatlist) beat-length)))
           no-of-beats))
    (list (list no-of-beats (/ 1 beat-length)) tree)))

(defun get-onsettime-before (timepoint abs-rhythm)
  (car (member timepoint (reverse abs-rhythm)
	       :test #'(lambda (item value) (> item (abs value))))))

(defun buildmeasure-seq (abs-rhythms timesigns)
  (let ((measure-start-points (om::dx->x 1 (mapcar #'(lambda (timesign)
						       (apply '/ timesign))
						   timesigns))))
    (loop for measure from 0 to (1- (length timesigns))
          collect (let ((this-seq (filter-events-between (nth measure measure-start-points) (nth (1+ measure) measure-start-points) abs-rhythms))
                        (this-timesign (nth measure timesigns))
                        local-onset)
                    ;;check if measure starts with tied pause
                    (if (and this-seq
                             (/= (abs (first this-seq)) (nth measure measure-start-points))
                             (> 0 (get-onsettime-before (nth measure measure-start-points) abs-rhythms)))
			(setf this-seq (append (list (- 0 (nth measure measure-start-points)))
					       this-seq)))
                    (if (and (not this-seq) 
                             (> 0 (get-onsettime-before (nth measure measure-start-points) abs-rhythms)))
			(setf this-seq (list (- 0 (nth measure measure-start-points)))))
                    (setf local-onset (build-local-times this-seq (nth measure measure-start-points)))
                    
                    (build-one-measure local-onset 
                                       (car this-timesign)
                                       (/ 1 (cadr this-timesign)))))))

(defun simple->tree (rhythmseq timesignseq)
  (let ((abs-rhythms (dx-x-pause-ok 1 (append rhythmseq '(-100)))))
    (list '? (buildmeasure-seq abs-rhythms timesignseq))))

(defun better-predefined-subdiv? (sub-tree)
  (let* ((proportional-list (cadr sub-tree))
	 (pauses (mapcar #'(lambda (value)
			     (if (< value 0) -1 1)) proportional-list))
	 (abs-proportional-list (mapcar 'abs proportional-list))
	 abs-answer)
    (setf abs-answer
          (cond ((equal abs-proportional-list '(2 2 2 3 3))
                 (list (list 2 (list (first pauses) (second pauses)
				     (third pauses)))
		       (fourth pauses)(fifth pauses))) ; added beginning parenthesis - NK
                ((equal abs-proportional-list '(3 3 2 2 2))
                 (list (first pauses) (second pauses)
		       (list 2 (list (third pauses) (fourth pauses)
				     (fifth pauses)))))
                ((equal abs-proportional-list '(3 2 2 2 3))
                 (list (first pauses) (list 2 (list (second pauses)
						    (third pauses)
						    (fourth pauses)))
		       (fifth pauses)))
                ((equal abs-proportional-list '(3.0 2 2 2 3))
                 (list (coerce (first pauses) 'float)
		       (list 2 (list (second pauses) (third pauses)
				     (fourth pauses)))
		       (fifth pauses)))
                ((equal abs-proportional-list '(3 3 4 2))
                 (list (first pauses) (second pauses)
		       (list 2 (list (* 2 (third pauses)) (fourth pauses)))))
                ((equal abs-proportional-list '(4 2 3 3))
                 (list (list 2 (list (* 2 (first pauses)) (second pauses)))
		       (third pauses) (fourth pauses)))
                ((equal abs-proportional-list '(2 4 3 3))
                 (list (list 2 (list (first pauses) (* 2 (second pauses))))
		       (third pauses) (fourth pauses)))
                ((equal abs-proportional-list '(3 3 2 4))
                 (list (first pauses) (second pauses)
		       (list 2 (list (third pauses) (* 2 (fourth pauses))))))
                ((equal abs-proportional-list '(3 1 1 1))
                 (list (first pauses) (list 1 (list (second pauses)
						    (third pauses)
						    (fourth pauses)))))
                ((equal abs-proportional-list '(3.0 1 1 1))
                 (list (coerce (first pauses) 'float)
		       (list 1 (list (second pauses) (third pauses)
				     (fourth pauses)))))
                ((equal abs-proportional-list '(1 1 1 3))
                 (list (list 1 (list (first pauses) (second pauses)
				     (third pauses))) (fourth pauses)))
                (t proportional-list)))
  (list 1 abs-answer)))

;;=============================================================================
;; om function
;;=============================================================================

;;little correction

(defun get-ratio-duration (l)
  (loop for item in l sum (abs item)))

(defmethod mktree ((rhythm list) (timesigns list)) ;; changed from om::defmethod - NK
		:initvals '((1/4 1/4 1/4 1/4) (4 4))
		:indoc '("list of integer ratio" "list of time signatures")
		:doc "Builds a hierarchical rhythm-tree from a simple list of note values.
The rhythm tree is intended for the tree input of a 'voice' factory box.

Inputs :

rhythm : a list of integer ratios. 1/4 is the quarter note.
timesigns : a list of time signatures, e.g. ( (4 4) (3 4) (5 8) ... )
If a single time signature (e.g. (4 4)) is given, it is extended as much as required
by the 'rhythm' length.

output:

A rhythm tree.
"
		:icon 254
  
		(if (typep (car timesigns) 'list)
		    (simple->tree  rhythm timesigns)
		  (simple->tree rhythm (make-list (1+ (truncate (* (/ (get-ratio-duration rhythm) (car timesigns)) (cadr timesigns)))) :initial-element timesigns))))

;;-----------------------------------------------------------------------------

(defun find-common-multiplier (multipliers)
  (apply 'lcm (mapcar #'(lambda (value) (/ 1 value)) multipliers)))

(defun make-multiplier-list (multipliers)
  (om* (find-common-multiplier multipliers) multipliers))

(defun fuse-ties (proportion-list)
  (let ((result nil))
    (loop while proportion-list
          do (progn (cond ((and result 
                                (typep (car proportion-list) 'float) 
                                (> (car proportion-list) 0))
                           (setf (car (last result)) (+ (car (last result)) (truncate (car proportion-list)))))
                          ((and result 
                                (typep (car proportion-list) 'float) 
				;;to avoid error for "slured pauses"
                                (< (car proportion-list) 0)) 
                           (setf (car proportion-list)
				 (truncate (car proportion-list)))
                           (setf result
				 (append result (list (car proportion-list)))))
                          (t 
                           (setf result
				 (append result
					 (list (car proportion-list))))))
                    (pop proportion-list)))
    result))

(defmethod sum-one-beat ((proportions number)) 1)

(defmethod sum-one-beat ((proportions list))
  (truncate (apply '+ (mapcar 'abs proportions))))

(defmethod unpack-tree ((sub-tree number)) sub-tree)

(defmethod unpack-tree ((sub-tree list))
  (let* ((proportional-format (mapcar 'unpack-tree (second sub-tree)))
         (prop-length-beat (mapcar 'sum-one-beat proportional-format))
         (multiples-for-list (mapcar #'(lambda (beat)
					 (/ (apply 'lcm prop-length-beat)
					    beat))
                                     prop-length-beat)))
    (om::flat (mapcar 'om::om* multiples-for-list proportional-format))))

;; ??????????
;; (defun measures->simple (measures)
;;   (let* ((proportions (mapcar #'(lambda (measure)
;;                                   (unpack-tree (list (caar measure) (cadr measure))))
;;                               measures))
;;          (multipliers (mapcar #'(lambda (measure proportion) (/ (apply '/ (car measure))
;;                                                      (truncate (apply '+ (mapcar 'abs proportion)))))
;;                               measures proportions))
;;          (global-proportion-list (flat (om* proportions (make-multiplier-list multipliers))))
;;          (global-multiplier (/ (first multipliers) (first (make-multiplier-list multipliers)))))


;;     (om* (fuse-ties (fuse-pauses global-proportion-list)) global-multiplier)
;;     ))

(defun measures->simple (measures)
  (let* ((proportions (mapcar #'(lambda (measure)
                                  (unpack-tree (list (caar measure)
                                                     (cadr measure))))
                              measures))
         (multipliers (mapcar #'(lambda (measure proportion)
				  (/ (apply
				      '/ (car measure))
				     (truncate (apply '+ (mapcar 'abs proportion)))))
                              measures proportions))
         (global-proportion-list (flat (om* proportions (make-multiplier-list multipliers))))
         (global-multiplier (/ (first multipliers)
			       (first (make-multiplier-list multipliers)))))
    (om* (fuse-pauses (fuse-ties global-proportion-list))
         global-multiplier)))

(defmethod tree2ratio ((tree list)) ;; changed from om::defmethod! - NK
		:initvals '('(2 (((4 4) (1 (1 (1 -2 1 1)) 1 1)) ((4 4) (1 (1 (1 2 1 1)) -1 1)))))
		:indoc '("tree")
		:doc "Convert a rhythm tree to a list of note values (ratios).

<tree> is a rhythm tree.

Pauses following each other will be fused to one (longer) 
pause. Slured note values will be fused to one note value."
		:icon 254
		
		(measures->simple (second tree)))

;;-EOF-------------------------------------------------------------------------


