;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output in Finale format from Common Music
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:30>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :cm-etf)

(defun eof-p (stream)
  "Return T if the stream has no more data in it"
  (null (peek-char nil stream nil nil)))

(defun eol-type (string)
  (cond ((not (find #\newline string)) 'mac)
        ((search (concatenate 'string
                              (string #\return)
                              (string #\newline))
                 string) 'dos)
        (t 'unix)))

(defun string-mac->unix (string)
  "Destructively substitutes newlines in place of returns in a string"
  (nsubstitute #\newline #\return string))

(defun string-dos->unix (string)
  "Destructively deletes returns in a string"
  (delete #\return string))

(defun string-unix->mac (string)
  "Destructively substitutes returns in place of newlines in a string"
  (nsubstitute #\return #\newline string))

(defun string-dos->mac (string)
  "Destructively deletes returns in a string"
  (delete #\newline string))

(defun string-unix->dos (string)
  "e.g. (string-unix->dos *new-etf*)"
  (with-output-to-string (output-stream)
                         (with-input-from-string (input-stream string)
                           (loop do (format output-stream 
                                            "~a~c~c" 
                                            (read-line input-stream)
                                            #\Return #\Newline)
                             until (eof-p input-stream)))))

(defun string-mac->dos (string)
  (string-unix->dos (string-mac->unix string)))

(defun string-unixify (string)
  (cond ((not (find #\newline string))                ; mac line endings
         (nsubstitute #\newline #\return string))
        ((search (format nil "~C~%" #\return) string) ; dos line endings
         (delete #\return string))
        (t string)))

(defun string-macify (string)
  (cond ((search (format nil "~C~%" #\return) string) ; dos line endings
         (remove #\newline string))
        ((find #\newline string)
         (substitute #\return #\newline string)) ;unix line endings
        (t string)))

(defun string-dosify (string)
  (cond ((search (format nil "~C~%" #\return) string) ; dos line endings
         string)
        ((find #\newline string)        ; unix line endings
         (string-unix->dos string))
        ((find #\return string)         ; mac line endings
         (string-mac->dos string))))

(defun read-raw-text (filename)
  (with-open-file (file-stream filename :direction :input 
                               #+clisp :external-format 
                               #+clisp (ext:make-encoding :charset 'charset:iso-8859-1 :line-terminator :unix)
			       #+sbcl :external-format
			       #+sbcl :ISO-8859-1)
     (with-output-to-string 
       (string-stream)
       (read-sequence (make-string (file-length file-stream))
                      (make-echo-stream file-stream string-stream)))))

;; (read-raw-text "/Users/Emily/Applications/cm-etf/template.etf")

(defun read-text (filename)
  (string-unixify (read-raw-text filename)))

(defun write-text (string filename &optional convert-fun)
  (let ((external-format
         #+sbcl :latin-1
         #+clisp (ext:make-encoding :charset 'charset:iso-8859-1 :line-terminator :unix)
         #-(or clisp sbcl) :default))
    (with-open-file (stream filename
                            :direction :output
                            :if-exists :supersede
                            :external-format external-format)
      (write-string (if (null convert-fun)
                        string
                        (funcall convert-fun string))
                    stream))
    (format t "~&wrote ~S" (namestring filename))
    (values)))
